import pytest

from gpaw import GPAW
from gpaw.tddft import TDDFT, DipoleMomentWriter, photoabsorption_spectrum
from gpaw.tddft.abc import PML, LinearAbsorbingBoundary, P4AbsorbingBoundary


@pytest.mark.later
def test_tddft_td_na2(in_tmp_dir, gpw_files):
    """Sodium dimer, Na2."""
    atoms = GPAW(gpw_files['na2_fd']).atoms
    # 16 fs run with 8.0 attosec time step
    time_step = 8.0  # 8.0 as (1 as = 0.041341 autime)5D
    iters = 10     # 2000 x 8 as => 16 fs
    # Weak delta kick to z-direction
    kick = [0, 0, 1e-3]

    # TDDFT calculator
    td_calc = TDDFT(gpw_files['na2_fd'])
    DipoleMomentWriter(td_calc, 'na2_dmz.dat')
    # Kick
    td_calc.absorption_kick(kick)
    # Propagate
    td_calc.propagate(time_step, iters)
    td_calc.write('na2_td.gpw', mode='all')
    # Linear absorption spectrum
    photoabsorption_spectrum('na2_dmz.dat', 'na2_spectrum_z.dat', width=0.3)

    iters = 3

    # test restart
    td_rest = TDDFT('na2_td.gpw')
    DipoleMomentWriter(td_rest, 'na2_dmz.dat')
    td_rest.propagate(time_step, iters)

    # test restart
    td_rest = TDDFT('na2_td.gpw', solver='BiCGStab')
    DipoleMomentWriter(td_rest, 'na2_dmz3.dat', force_new_file=True)
    td_rest.propagate(time_step, iters)

    # test absorbing boundary conditions

    # linear imaginary potential
    td_ipabs = TDDFT(gpw_files['na2_fd'])
    ip_abc = LinearAbsorbingBoundary(5.0, 0.01, atoms.positions)
    td_ipabs.set_absorbing_boundary(ip_abc)
    DipoleMomentWriter(td_ipabs, 'na2_dmz4.dat')
    td_ipabs.propagate(time_step, iters)

    # 4th order polynomial (1-(x^2-1)^2) imaginary potential
    td_ip4abs = TDDFT(gpw_files['na2_fd'])
    ip4_abc = P4AbsorbingBoundary(5.0, 0.03, atoms.positions, 3.0)
    td_ip4abs.set_absorbing_boundary(ip4_abc)
    DipoleMomentWriter(td_ip4abs, 'na2_dmz5.dat')
    td_ip4abs.propagate(time_step, iters)

    # perfectly matched layers
    td_pmlabs = TDDFT(gpw_files['na2_fd'], solver='BiCGStab')
    pml_abc = PML(100.0, 0.1)
    td_pmlabs.set_absorbing_boundary(pml_abc)
    DipoleMomentWriter(td_pmlabs, 'na2_dmz6.dat')
    td_pmlabs.propagate(time_step, iters)


@pytest.mark.later
def test_tddft_fail_with_symmetry(in_tmp_dir, gpw_files):

    # Time-propagation calculation
    # should not be allowed with symmetries
    with pytest.raises(ValueError):
        TDDFT(gpw_files['na2_fd_with_sym'])



import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft.spectrum import spectrum


@pytest.mark.lrtddft
def test_lrtddft_2(in_tmp_dir):
    txt = '-'
    txt = None
    xc = 'LDA'

    R = 0.7  # approx. experimental bond length
    a = 4.0
    c = 5.0
    H2 = Atoms('HH',
               [(a / 2, a / 2, (c - R) / 2),
                (a / 2, a / 2, (c + R) / 2)],
               cell=(a, a, c))

    calc = GPAW(mode='fd',
                xc=xc,
                nbands=2,
                spinpol=False,
                eigensolver='rmm-diis',
                txt=txt)
    H2.calc = calc
    H2.get_potential_energy()
    calc.write('H2saved_wfs.gpw', 'all')
    calc.write('H2saved.gpw')

    print('-> starting directly after a gs calculation')
    lr = LrTDDFT(calc, txt='-')
    lr.diagonalize()

    print('-> reading gs with wfs')
    gs = GPAW('H2saved_wfs.gpw', txt=txt)

    lr1 = LrTDDFT(gs, xc=xc, txt='-')
    # check the oscillator strrength
    assert (abs(lr1[0].get_oscillator_strength()[0] /
                lr[0].get_oscillator_strength()[0] - 1) < 1e-7)

    print('-> reading gs without wfs')
    gs = GPAW('H2saved.gpw', txt=None)

    lr2 = LrTDDFT(gs, txt='-')
    # check the oscillator strrength
    d = abs(lr2[0].get_oscillator_strength()[0] /
            lr[0].get_oscillator_strength()[0] - 1)
    assert (d < 2e-3), d

    # write spectrum
    spectrum(lr, 'spectrum.dat')



import pickle
import numpy as np
from math import pi
import ase.units
import os

Hartree = ase.units.Hartree
Bohr = ase.units.Bohr


def load(fd):
    try:
        return pickle.load(fd, encoding='latin1')
    except TypeError:
        return pickle.load(fd)


class BuildingBlock:

    """ Module for using Linear response to calculate dielectric
    building block of 2D material with GPAW"""

    def __init__(self, filename, df, isotropic_q=True, nq_inf=10,
                 direction='x', qmax=None, txt='-'):
        """Creates a BuildingBlock object.

        filename: str
            used to save data file: filename-chi.npz
        df: DielectricFunction object
            Determines how linear response calculation is performed
        isotropic_q: bool
            If True, only q-points along one direction (1 0 0) in the
            2D BZ is included, thus assuming an isotropic material
        direction: 'x' or 'y'
            Direction used for isotropic q sampling.
        qmax: float
            Cutoff for q-grid. To be used if one wishes to sample outside the
            irreducible BZ. Only works for isotropic q-sampling.
        nq_inf: int
            number of extra q points in the limit q->0 along each direction,
            extrapolated from q=0, assuming that the head of chi0_wGG goes
            as q^2 and the wings as q.
            Note that this does not hold for (semi)metals!

        """
        assert isotropic_q, "Non-isotropic calculation" \
            + " temporarily turned-off until properly tested."
        if qmax is not None:
            assert isotropic_q
        self.filename = filename
        self.isotropic_q = isotropic_q
        self.nq_inf = nq_inf
        self.nq_inftot = nq_inf
        if not isotropic_q:
            self.nq_inftot *= 2

        if direction == 'x':
            qdir = 0
        elif direction == 'y':
            qdir = 1
        self.direction = direction

        self.df = df  # dielectric function object
        assert self.df.coulomb.truncation == '2D'
        self.wd = self.df.wd

        self.context = self.df.context.with_txt(txt)

        gs = self.df.gs
        kd = gs.kd
        self.kd = kd
        r = gs.gd.get_grid_point_coordinates()
        self.z = r[2, 0, 0, :]

        nw = len(self.wd)
        self.chiM_qw = np.zeros([0, nw])
        self.chiD_qw = np.zeros([0, nw])
        self.drhoM_qz = np.zeros([0, self.z.shape[0]])
        self.drhoD_qz = np.zeros([0, self.z.shape[0]])

        # First: choose all ibzq in 2D BZ
        from ase.dft.kpoints import monkhorst_pack
        from gpaw.kpt_descriptor import KPointDescriptor
        offset_c = 0.5 * ((kd.N_c + 1) % 2) / kd.N_c
        bzq_qc = monkhorst_pack(kd.N_c) + offset_c
        qd = KPointDescriptor(bzq_qc)
        qd.set_symmetry(gs.atoms, kd.symmetry)
        q_cs = qd.ibzk_kc
        rcell_cv = 2 * pi * np.linalg.inv(gs.gd.cell_cv).T
        if isotropic_q:  # only use q along [1 0 0] or [0 1 0] direction.
            Nk = kd.N_c[qdir]
            qx = np.array(range(1, Nk // 2)) / float(Nk)
            q_cs = np.zeros([Nk // 2 - 1, 3])
            q_cs[:, qdir] = qx
            q = 0
            if qmax is not None:
                qmax *= Bohr
                qmax_v = np.zeros([3])
                qmax_v[qdir] = qmax
                q_c = q_cs[-1]
                q_v = np.dot(q_c, rcell_cv)
                q = (q_v**2).sum()**0.5
                assert Nk % 2 == 0
                i = Nk / 2.0
                while q < qmax:
                    if i == Nk:  # omit BZ edge
                        i += 1
                        continue
                    q_c = np.zeros([3])
                    q_c[qdir] = i / Nk
                    q_cs = np.append(q_cs, q_c[np.newaxis, :], axis=0)
                    q_v = np.dot(q_c, rcell_cv)
                    q = (q_v**2).sum()**0.5
                    i += 1
        q_vs = np.dot(q_cs, rcell_cv)
        q_abs = (q_vs**2).sum(axis=1)**0.5
        sort = np.argsort(q_abs)
        q_abs = q_abs[sort]
        q_cs = q_cs[sort]
        if isotropic_q:
            q_cut = q_abs[0] / 2  # extrapolate to half of smallest finite q
        else:
            q_cut = q_abs[1]  # smallest finite q
        self.nq_cut = self.nq_inftot + 1

        q_infs = np.zeros([q_cs.shape[0] + self.nq_inftot, 3])
        # x-direction:
        q_infs[: self.nq_inftot, qdir] = \
            np.linspace(1e-05, q_cut, self.nq_inftot + 1)[:-1]
        if not isotropic_q:  # y-direction
            q_infs[self.nq_inf:self.nq_inftot, 1] = \
                np.linspace(0, q_cut, self.nq_inf + 1)[1:]

        # add q_inf to list
        self.q_cs = np.insert(q_cs, 0, np.zeros([self.nq_inftot, 3]), axis=0)
        self.q_vs = np.dot(self.q_cs, rcell_cv)
        self.q_vs += q_infs
        self.q_abs = (self.q_vs**2).sum(axis=1)**0.5
        self.q_infs = q_infs
        self.complete = False
        self.nq = 0
        if self.load_chi_file():
            if self.complete:
                self.context.print('Building block loaded from file')
        self.context.comm.barrier()

    def calculate_building_block(self, add_intraband=False):
        if self.complete:
            return
        Nq = self.q_cs.shape[0]
        for nq in range(self.nq, Nq):
            self.nq = nq
            self.save_chi_file()
            q_c = self.q_cs[nq]
            q_inf = self.q_infs[nq]
            if np.allclose(q_inf, 0):
                q_inf = None

            qcstr = '(' + ', '.join(['%.3f' % x for x in q_c]) + ')'
            self.context.print(
                'Calculating contribution from q-point #%d/%d, q_c=%s' % (
                    nq + 1, Nq, qcstr), flush=False)
            if q_inf is not None:
                qstr = '(' + ', '.join(['%.3f' % x for x in q_inf]) + ')'
                self.context.print('    and q_inf=%s' % qstr, flush=False)
            qpd, chi0_wGG, \
                chi_wGG = self.df.get_dielectric_matrix(
                    symmetric=False,
                    calculate_chi=True,
                    q_c=q_c,
                    q_v=q_inf,
                    direction=self.direction,
                    add_intraband=add_intraband)
            self.context.print('calculated chi!')

            nw = len(self.wd)
            comm = self.context.comm
            w1 = min(self.df.blocks1d.blocksize * comm.rank, nw)

            _, _, chiM_qw, chiD_qw, _, drhoM_qz, drhoD_qz = \
                get_chi_2D(self.wd.omega_w, qpd, chi_wGG)

            chiM_w = chiM_qw[0]
            chiD_w = chiD_qw[0]
            chiM_w = self.collect(chiM_w)
            chiD_w = self.collect(chiD_w)

            if self.context.comm.rank == 0:
                assert w1 == 0  # drhoM and drhoD in static limit
                self.update_building_block(chiM_w[np.newaxis, :],
                                           chiD_w[np.newaxis, :],
                                           drhoM_qz, drhoD_qz)

        # Induced densities are not probably described in q-> 0 limit-
        # replace with finite q result:
        if self.context.comm.rank == 0:
            for n in range(Nq):
                if np.allclose(self.q_cs[n], 0):
                    self.drhoM_qz[n] = self.drhoM_qz[self.nq_cut]
                    self.drhoD_qz[n] = self.drhoD_qz[self.nq_cut]

        self.complete = True
        self.save_chi_file()

        return

    def update_building_block(self, chiM_qw, chiD_qw, drhoM_qz,
                              drhoD_qz):

        self.chiM_qw = np.append(self.chiM_qw, chiM_qw, axis=0)
        self.chiD_qw = np.append(self.chiD_qw, chiD_qw, axis=0)
        self.drhoM_qz = np.append(self.drhoM_qz, drhoM_qz, axis=0)
        self.drhoD_qz = np.append(self.drhoD_qz, drhoD_qz, axis=0)

    def save_chi_file(self, filename=None):
        if filename is None:
            filename = self.filename
        data = {'last_q': self.nq,
                'complete': self.complete,
                'isotropic_q': self.isotropic_q,
                'q_cs': self.q_cs,
                'q_vs': self.q_vs,
                'q_abs': self.q_abs,
                'omega_w': self.wd.omega_w,
                'chiM_qw': self.chiM_qw,
                'chiD_qw': self.chiD_qw,
                'z': self.z,
                'drhoM_qz': self.drhoM_qz,
                'drhoD_qz': self.drhoD_qz}

        if self.context.comm.rank == 0:
            np.savez_compressed(filename + '-chi.npz',
                                **data)
        self.context.comm.barrier()

    def load_chi_file(self):
        try:
            data = np.load(self.filename + '-chi.npz')
        except IOError:
            return False
        if (np.all(data['omega_w'] == self.wd.omega_w) and
            np.all(data['q_cs'] == self.q_cs) and
            np.all(data['z'] == self.z)):
            self.nq = data['last_q']
            self.complete = data['complete']
            self.chiM_qw = data['chiM_qw']
            self.chiD_qw = data['chiD_qw']
            self.drhoM_qz = data['drhoM_qz']
            self.drhoD_qz = data['drhoD_qz']
            return True
        else:
            return False

    def interpolate_to_grid(self, q_grid, w_grid):

        """
        Parameters
        q_grid: in Ang. should start at q=0
        w_grid: in eV
        """

        from scipy.interpolate import RectBivariateSpline
        from scipy.interpolate import interp1d
        from gpaw.response.frequencies import FrequencyGridDescriptor
        if not self.complete:
            self.calculate_building_block()
        q_grid *= Bohr
        w_grid /= Hartree

        assert np.max(q_grid) <= np.max(self.q_abs), \
            'q can not be larger that %1.2f Ang' % np.max(self.q_abs / Bohr)
        assert np.max(w_grid) <= np.max(self.wd.omega_w), \
            'w can not be larger that %1.2f eV' % \
            np.max(self.wd.omega_w * Hartree)

        sort = np.argsort(self.q_abs)
        q_abs = self.q_abs[sort]

        # chi monopole
        self.chiM_qw = self.chiM_qw[sort]

        omit_q0 = False
        if np.isclose(q_abs[0], 0) and not np.isclose(self.chiM_qw[0, 0], 0):
            omit_q0 = True  # omit q=0 from interpolation
            q0_abs = q_abs[0].copy()
            q_abs[0] = 0.
            chi0_w = self.chiM_qw[0].copy()
            self.chiM_qw[0] = np.zeros_like(chi0_w)

        yr = RectBivariateSpline(q_abs, self.wd.omega_w,
                                 self.chiM_qw.real,
                                 s=0)

        yi = RectBivariateSpline(q_abs, self.wd.omega_w,
                                 self.chiM_qw.imag, s=0)

        self.chiM_qw = yr(q_grid, w_grid) + 1j * yi(q_grid, w_grid)
        if omit_q0:
            yr = interp1d(self.wd.omega_w, chi0_w.real)
            yi = interp1d(self.wd.omega_w, chi0_w.imag)
            chi0_w = yr(w_grid) + 1j * yi(w_grid)
            q_abs[0] = q0_abs
            if np.isclose(q_grid[0], 0):
                self.chiM_qw[0] = chi0_w

        # chi dipole
        yr = RectBivariateSpline(q_abs, self.wd.omega_w,
                                 self.chiD_qw[sort].real,
                                 s=0)
        yi = RectBivariateSpline(q_abs, self.wd.omega_w,
                                 self.chiD_qw[sort].imag,
                                 s=0)

        self.chiD_qw = yr(q_grid, w_grid) + 1j * yi(q_grid, w_grid)

        # drho monopole

        yr = RectBivariateSpline(q_abs, self.z,
                                 self.drhoM_qz[sort].real, s=0)
        yi = RectBivariateSpline(q_abs, self.z,
                                 self.drhoM_qz[sort].imag, s=0)

        self.drhoM_qz = yr(q_grid, self.z) + 1j * yi(q_grid, self.z)

        # drho dipole
        yr = RectBivariateSpline(q_abs, self.z,
                                 self.drhoD_qz[sort].real, s=0)
        yi = RectBivariateSpline(q_abs, self.z,
                                 self.drhoD_qz[sort].imag, s=0)

        self.drhoD_qz = yr(q_grid, self.z) + 1j * yi(q_grid, self.z)

        self.q_abs = q_grid
        self.wd = FrequencyGridDescriptor(w_grid)
        self.save_chi_file(filename=self.filename + '_int')

    def collect(self, a_w):
        comm = self.context.comm
        mynw = self.df.blocks1d.blocksize
        b_w = np.zeros(mynw, a_w.dtype)
        b_w[:self.df.blocks1d.nlocal] = a_w
        nw = len(self.wd)
        A_w = np.empty(comm.size * mynw, a_w.dtype)
        comm.all_gather(b_w, A_w)
        return A_w[:nw]

    def clear_temp_files(self):
        if not self.savechi0:
            comm = self.context.comm
            if comm.rank == 0:
                while len(self.temp_files) > 0:
                    filename = self.temp_files.pop()
                    os.remove(filename)


"""TOOLS"""


def check_building_blocks(BBfiles=None):
    """ Check that building blocks are on same frequency-
    and q- grid.

    BBfiles: list of str
        list of names of BB files
    """
    name = BBfiles[0] + '-chi.npz'
    data = np.load(name)
    try:
        q = data['q_abs'].copy()
        w = data['omega_w'].copy()
    except TypeError:
        # Skip test for old format:
        return True
    for name in BBfiles[1:]:
        data = np.load(name + '-chi.npz')
        if len(w) != len(data['omega_w']):
            return False
        elif not ((data['q_abs'] == q).all() and
                  (data['omega_w'] == w).all()):
            return False
    return True


def get_chi_2D(omega_w=None, qpd=None, chi_wGG=None, q0=None,
               filenames=None, name=None):
    r"""Calculate the monopole and dipole contribution to the
    2D susceptibillity chi_2D, defined as

    ::

      \chi^M_2D(q, \omega) = \int\int dr dr' \chi(q, \omega, r,r') \\
                          = L \chi_{G=G'=0}(q, \omega)
      \chi^D_2D(q, \omega) = \int\int dr dr' z \chi(q, \omega, r,r') z'
                           = 1/L sum_{G_z,G_z'} z_factor(G_z)
                           chi_{G_z,G_z'} z_factor(G_z'),
      Where z_factor(G_z) =  +/- i e^{+/- i*G_z*z0}
      (L G_z cos(G_z L/2)-2 sin(G_z L/2))/G_z^2

    input parameters:

    filenames: list of str
        list of chi_wGG.pckl files for different q calculated with
        the DielectricFunction module in GPAW
    name: str
        name writing output files
    """

    q_list_abs = []
    if chi_wGG is None and filenames is not None:
        omega_w, qpd, chi_wGG, q0 = read_chi_wGG(filenames[0])
        nq = len(filenames)
    elif chi_wGG is not None:
        nq = 1
    nw = chi_wGG.shape[0]
    r = qpd.gd.get_grid_point_coordinates()
    z = r[2, 0, 0, :]
    L = qpd.gd.cell_cv[2, 2]  # Length of cell in Bohr
    z0 = L / 2.  # position of layer
    chiM_qw = np.zeros([nq, nw], dtype=complex)
    chiD_qw = np.zeros([nq, nw], dtype=complex)
    drhoM_qz = np.zeros([nq, len(z)], dtype=complex)  # induced density
    drhoD_qz = np.zeros([nq, len(z)], dtype=complex)  # induced dipole density

    for iq in range(nq):
        if iq != 0:
            omega_w, qpd, chi_wGG, q0 = read_chi_wGG(filenames[iq])
        if q0 is not None:
            q = q0
        else:
            q = qpd.K_qv
        npw = chi_wGG.shape[1]
        G_Gv = qpd.get_reciprocal_vectors(add_q=False)

        Glist = []
        for iG in range(npw):  # List of G with Gx,Gy = 0
            if G_Gv[iG, 0] == 0 and G_Gv[iG, 1] == 0:
                Glist.append(iG)
        q_abs = np.linalg.norm(q)
        q_list_abs.append(q_abs)

        # If node lacks frequency points due to block parallelization then
        # return empty arrays
        if nw == 0:
            continue
        chiM_qw[iq] = L * chi_wGG[:, 0, 0]
        drhoM_qz[iq] += chi_wGG[0, 0, 0]
        for iG in Glist[1:]:
            G_z = G_Gv[iG, 2]
            qGr_R = np.inner(G_z, z.T).T
            # Fourier transform to get induced density at \omega=0
            drhoM_qz[iq] += np.exp(1j * qGr_R) * chi_wGG[0, iG, 0]
            for iG1 in Glist[1:]:
                G_z1 = G_Gv[iG1, 2]
                # integrate with z along both coordinates
                factor = z_factor(z0, L, G_z)
                factor1 = z_factor(z0, L, G_z1, sign=-1)
                chiD_qw[iq, :] += 1. / L * factor * chi_wGG[:, iG, iG1] * \
                    factor1
                # induced dipole density due to V_ext = z
                drhoD_qz[iq, :] += 1. / L * np.exp(1j * qGr_R) * \
                    chi_wGG[0, iG, iG1] * factor1
    # Normalize induced densities with chi
    if nw != 0:
        drhoM_qz /= np.repeat(chiM_qw[:, 0, np.newaxis], drhoM_qz.shape[1],
                              axis=1)
        drhoD_qz /= np.repeat(chiD_qw[:, 0, np.newaxis], drhoM_qz.shape[1],
                              axis=1)

    """ Returns q array, frequency array, chi2D monopole and dipole, induced
    densities and z array (all in Bohr)
    """
    if name is not None:
        arrays = [np.array(q_list_abs), omega_w, chiM_qw, chiD_qw,
                  z, drhoM_qz, drhoD_qz]
        names = ['q_abs', 'omega_w', 'z',
                 'chiM_qw', 'chiD_qw',
                 'drhoM_qz', 'drhoD_qz']
        data = {}
        for array, name in zip(arrays, names):
            data[name] = array
        np.save(name + '-chi.npz', **data)
    return np.array(q_list_abs) / Bohr, omega_w * Hartree, chiM_qw, \
        chiD_qw, z, drhoM_qz, drhoD_qz


def z_factor(z0, d, G, sign=1):
    factor = -1j * sign * np.exp(1j * sign * G * z0) * \
        (d * G * np.cos(G * d / 2.) - 2. * np.sin(G * d / 2.)) / G**2
    return factor


def z_factor2(z0, d, G, sign=1):
    factor = sign * np.exp(1j * sign * G * z0) * np.sin(G * d / 2.)
    return factor


def expand_layers(structure):
    newlist = []
    for name in structure:
        num = ''
        while name[0].isdigit():
            num += name[0]
            name = name[1:]
        try:
            num = int(num)
        except ValueError:
            num = 1
        for n in range(num):
            newlist.append(name)
    return newlist


def read_chi_wGG(name):
    """
    Read density response matrix calculated with the DielectricFunction
    module in GPAW.
    Returns frequency grid, gpaw.wavefunctions object, chi_wGG
    """
    fd = open(name, 'rb')
    omega_w, qpd, chi_wGG, q0, chi0_wvv = load(fd)
    nw = len(omega_w)
    nG = qpd.ngmax
    chi_wGG = np.empty((nw, nG, nG), complex)
    for chi_GG in chi_wGG:
        chi_GG[:] = load(fd)
    return omega_w, qpd, chi_wGG, q0



import pytest

import numpy as np

from gpaw.new.ase_interface import GPAW
from gpaw.spinorbit import soc_eigenstates


def test_orbmag_Ni(gpw_files):
    # Parameters

    easy_axis = 1 / np.sqrt(3) * np.ones(3)
    theta = np.rad2deg(np.arccos(easy_axis[2]))
    phi = 45

    # Collinear calculation

    calc_col = GPAW(gpw_files['fcc_Ni_col'],
                    parallel={'domain': 1, 'band': 1})

    energy_col = calc_col.get_potential_energy(calc_col.atoms)
    density = calc_col.calculation.state.density
    magmoms_col_v, _ = density.calculate_magnetic_moments()
    with pytest.raises(AssertionError, match='Collinear calculations*'):
        calc_col.get_orbital_magnetic_moments()
    orbmag_col_v = soc_eigenstates(calc_col,
                                   theta=theta,
                                   phi=phi).get_orbital_magnetic_moments()[0]

    # Non-collinear calculation without self-consistent spin-orbit

    calc_ncol = GPAW(gpw_files['fcc_Ni_ncol'],
                     parallel={'domain': 1, 'band': 1})

    energy_ncol = calc_ncol.get_potential_energy(calc_ncol.atoms)
    density = calc_ncol.calculation.state.density
    magmoms_ncol_v, _ = density.calculate_magnetic_moments()
    orbmag_ncol_v = soc_eigenstates(
        calc_ncol).get_orbital_magnetic_moments()[0]

    # Test that col and ncol give the same groundstate (with rotated magmoms)
    # and the same orbital magnetic moments from the soc_eigenstates module

    dif_energy = energy_ncol - energy_col
    dif_magmom = np.linalg.norm(magmoms_ncol_v) - magmoms_col_v[2]
    dif_orbmag = np.linalg.norm(orbmag_ncol_v - orbmag_col_v)

    assert dif_energy == pytest.approx(0, abs=1.0e-6)
    assert dif_magmom == pytest.approx(0, abs=1.0e-6)
    assert dif_orbmag == pytest.approx(0, abs=1.0e-6)

    # Non-collinear calculation with self-consistent spin-orbit

    calc_ncolsoc = GPAW(gpw_files['fcc_Ni_ncolsoc'],
                        parallel={'domain': 1, 'band': 1})

    calc_ncolsoc.get_potential_energy(calc_ncolsoc.atoms)
    orbmag_ncolsoc_v = calc_ncolsoc.get_orbital_magnetic_moments()[0]

    # Assert direction and magnitude of orbital magnetic moment
    assert np.linalg.norm(orbmag_ncolsoc_v) == pytest.approx(
        0.045176, abs=1e-6)
    assert np.dot(orbmag_ncolsoc_v, easy_axis) == pytest.approx(
        0.045176, abs=1e-6)

    # Get difference between orbital magnetic moments when soc is included
    # self-consistently. Assert that this difference doesn't change.

    dif_orbmag2 = np.linalg.norm(orbmag_ncolsoc_v - orbmag_col_v)
    dif_orbmag3 = np.linalg.norm(orbmag_ncolsoc_v - orbmag_ncol_v)

    assert dif_orbmag2 == pytest.approx(0.002516, abs=1e-6)
    assert dif_orbmag3 == pytest.approx(0.002516, abs=1e-6)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Domain object.

This module contains the definition of the ``Domain`` class and some
helper functins for parallel domain decomposition.  """

import numpy as np

UNIFORM = False  # distribute grid points uniformly


class Domain:
    """Domain class.

    A ``Domain`` object (in domain.py) holds informaion on the unit
    cell and the boundary conditions"""

    def __init__(self, cell, pbc, comm, parsize_c, N_c):
        """Create Domain object from a unit cell and boundary conditions.

        The arguments are the lengths of the three axes, followed by a
        tuple of three periodic-boundary condition flags (``bool``s).

        Parallel stuff:
         =============== ==================================================
         ``sdisp_cd``    Scaled displacement in direction ``d`` along axis
                         ``c``.
         ``comm``        MPI-communicator for this domain.
         ``neighbor_cd`` Rank of neighbor CPU in direction ``d`` along axis
                         ``c``.
         ``parpos_c``    Position of this CPU in the 3D grid of all CPUs.
         ``parsize_c``   Domain decomposition along the three axes.
         ``stride_c``    Strides.
         =============== ==================================================
        """
        cell_c = np.array(cell, float)
        if cell_c.ndim == 1:
            self.cell_cv = np.diag(cell_c)
        else:
            self.cell_cv = cell_c
            cell_c = (self.cell_cv**2).sum(1)**0.5
        self.icell_cv = np.linalg.inv(self.cell_cv).T
        self.xxxucell_cv = np.array([self.cell_cv[x] / cell_c[x]
                                     for x in range(3)])
        self.xxxiucell_cv = np.linalg.inv(self.xxxucell_cv.T)  # Jacobian

        self.pbc_c = np.asarray(pbc, bool)

        if isinstance(parsize_c, int):
            assert parsize_c == comm.size
            parsize_c = None

        self.set_decomposition(comm, parsize_c, N_c)

    def set_decomposition(self, comm, parsize_c=None, N_c=None):
        """Set MPI-communicator and do domain decomposition.

        With ``parsize_c=(a, b, c)``, the domin will be divided in
        ``a*b*c`` sub-domains - one for each CPU in ``comm``.  If
        ``parsize_c`` is not given, the number of grid points will be
        used to suggest a good domain decomposition."""

        self.comm = comm

        if parsize_c is None:
            # XXX This should rather be:
            # parsize_c = decompose_domain(N_c - not self.pbc_c, comm.size)
            parsize_c = decompose_domain(N_c, comm.size)
        self.parsize_c = np.array(parsize_c)

        self.stride_c = np.array([parsize_c[1] * parsize_c[2],
                                  parsize_c[2],
                                  1])

        if np.prod(self.parsize_c) != self.comm.size:
            raise RuntimeError('Bad domain decomposition! '
                               'CPU counts %s do not multiply to '
                               'communicator size %d' % (self.parsize_c,
                                                         self.comm.size))

        self.parpos_c = self.get_processor_position_from_rank()
        self.find_neighbor_processors()

    def get_ranks_from_positions(self, spos_ac):
        """Calculate rank of domain containing scaled position."""
        rnk_ac = np.floor(spos_ac * self.parsize_c).astype(int)
        if (rnk_ac < 0).any() or (rnk_ac >= self.parsize_c).any():
            raise ValueError('Some atom is too close to the zero-boundary!')
        return np.dot(rnk_ac, self.stride_c)

    def get_rank_from_position(self, spos_c):
        return self.get_ranks_from_positions(np.array([spos_c]))[0]

    def get_rank_from_processor_position(self, parpos_c):
        return parpos_c[2] + self.parsize_c[2] * \
            (parpos_c[1] + self.parsize_c[1] * parpos_c[0])

    def get_processor_position_from_rank(self, rank=None):
        """Calculate position of a domain in the 3D grid of all domains."""
        if rank is None:
            rank = self.comm.rank
        parpos_c = np.array([rank // self.stride_c[0],
                             (rank % self.stride_c[0]) // self.stride_c[1],
                             rank % self.stride_c[1]])
        assert np.dot(parpos_c, self.stride_c) == rank
        return parpos_c

    def find_neighbor_processors(self):
        """Find neighbor processors - surprise!

        With ``i`` and ``d`` being the indices for the axis (x, y or
        z) and direction + or - (0 or 1), two attributes are
        calculated:

        * ``neighbor_cd``:  Rank of neighbor.
        * ``sdisp_cd``:  Scaled displacement for neighbor.
        """

        self.neighbor_cd = np.zeros((3, 2), int)
        self.sdisp_cd = np.zeros((3, 2), int)
        for c in range(3):
            p = self.parpos_c[c]
            for d in range(2):
                sdisp = 2 * d - 1
                pd = p + sdisp
                pd0 = pd % self.parsize_c[c]
                self.neighbor_cd[c, d] = (self.comm.rank +
                                          (pd0 - p) * self.stride_c[c])
                if pd0 != pd:
                    # Wrap around the box?
                    if self.pbc_c[c]:
                        # Yes:
                        self.sdisp_cd[c, d] = -sdisp
                    else:
                        # No:
                        self.neighbor_cd[c, d] = -1


def decompose_domain(ngpts_c, ncores):
    """Determine parsize based on grid size and number of processors.

    If global variable UNIFORM is True, the returned parsize will
    result in a uniform distribution of gridpoints amongst processors.

    If UNIFORM is False, the returned parsize may result in a variable
    number of points on each cpu.
    """
    if ncores == 1:
        return (1, 1, 1)
    n1, n2, n3 = ngpts_c
    plist = prims(ncores)
    pdict = {}
    for n in plist:
        pdict[n] = 0
    for n in plist:
        pdict[n] += 1
    candidates = factorizations(list(pdict.items()))
    mincost = 10000000000.0
    best = None
    for p1, p2, p3 in candidates:
        if UNIFORM and (n1 % p1 != 0 or n2 % p2 != 0 or n3 % p3 != 0):
            continue
        elif UNIFORM:
            m1 = n1 // p1
            m2 = n2 // p2
            m3 = n3 // p3
        else:
            m1 = float(n1) / p1
            m2 = float(n2) / p2
            m3 = float(n3) / p3
        cost = abs(m1 - m2) + abs(m2 - m3) + abs(m3 - m1)
        # A long z-axis (unit stride) is best:
        if m1 <= m2 <= m3:
            cost -= 0.1
        if cost < mincost:
            mincost = cost
            best = (p1, p2, p3)
    if best is None:
        raise RuntimeError("Can't decompose a %dx%dx%d grid on %d CPUs!" %
                           (n1, n2, n3, ncores))
    return best


def factorizations(f):
    p, n = f[0]
    facs0 = []
    for n1 in range(n + 1):
        for n2 in range(n - n1 + 1):
            n3 = n - n1 - n2
            facs0.append((p**n1, p**n2, p**n3))
    if len(f) == 1:
        return facs0
    else:
        facs = factorizations(f[1:])
        facs1 = []
        for p1, p2, p3 in facs0:
            facs1 += [(p1 * q1, p2 * q2, p3 * q3) for q1, q2, q3 in facs]
        return facs1


primes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53,
          59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113,
          127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181,
          191, 193, 197, 199]


def prims(p):
    if p == 1:
        return []
    for n in primes:
        if p % n == 0:
            return prims(p / n) + [n]
    raise RuntimeError



from __future__ import annotations

from math import pi
from typing import Sequence

import numpy as np

import gpaw.fftw as fftw
from gpaw.core.arrays import DistributedArrays
from gpaw.core.atom_centered_functions import UGAtomCenteredFunctions
from gpaw.core.domain import Domain
from gpaw.gpu import as_xp
from gpaw.grid_descriptor import GridDescriptor
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import cached_property, zips
from gpaw.typing import (Array1D, Array2D, Array3D, Array4D, ArrayLike1D,
                         ArrayLike2D, Vector)
from gpaw.new.c import add_to_density, symmetrize_ft
from gpaw.fd_operators import Gradient


class UGDesc(Domain):
    def __init__(self,
                 *,
                 cell: ArrayLike1D | ArrayLike2D,  # bohr
                 size: ArrayLike1D,
                 pbc=(True, True, True),
                 kpt: Vector | None = None,  # in units of reciprocal cell
                 comm: MPIComm = serial_comm,
                 decomp: Sequence[Sequence[int]] | None = None,
                 dtype=None):
        """Description of 3D uniform grid.

        parameters
        ----------
        cell:
            Unit cell given as three floats (orthorhombic grid), six floats
            (three lengths and the angles in degrees) or a 3x3 matrix
            (units: bohr).
        size:
            Number of grid points along axes.
        pbc:
            Periodic boundary conditions flag(s).
        comm:
            Communicator for domain-decomposition.
        kpt:
            K-point for Block-boundary conditions specified in units of the
            reciprocal cell.
        decomp:
            Decomposition of the domain.
        dtype:
            Data-type (float or complex).
        """
        self.size_c = np.array(size, int)

        if decomp is None:
            gd = GridDescriptor(size, pbc_c=pbc, comm=comm)
            decomp = gd.n_cp
        self.decomp_cp = [np.asarray(d) for d in decomp]

        self.parsize_c = np.array([len(d_p) - 1 for d_p in self.decomp_cp])
        self.mypos_c = np.unravel_index(comm.rank, self.parsize_c)

        self.start_c = np.array([d_p[p]
                                 for d_p, p
                                 in zips(self.decomp_cp, self.mypos_c)])
        self.end_c = np.array([d_p[p + 1]
                               for d_p, p
                               in zips(self.decomp_cp, self.mypos_c)])
        self.mysize_c = self.end_c - self.start_c

        Domain.__init__(self, cell, pbc, kpt, comm, dtype)
        self.myshape = tuple(self.mysize_c)

        self.dv = self.volume / self.size_c.prod()

        self.itemsize = 8 if self.dtype == float else 16

    @property
    def size(self):
        """Size of uniform grid."""
        return self.size_c.copy()

    def global_shape(self) -> tuple[int, ...]:
        """Actual size of uniform grid."""
        return tuple(self.size_c - 1 + self.pbc_c)

    def __repr__(self):
        return Domain.__repr__(self).replace(
            'Domain(',
            f'UGDesc(size={self.size_c.tolist()}, ')

    def _short_string(self, global_shape):
        return f'uniform wave function grid shape: {global_shape}'

    @cached_property
    def phase_factor_cd(self):
        """Phase factor for block-boundary conditions."""
        delta_d = np.array([-1, 1])
        disp_cd = np.empty((3, 2))
        for pos, pbc, size, disp_d in zips(self.mypos_c, self.pbc_c,
                                           self.parsize_c, disp_cd):
            disp_d[:] = -((pos + delta_d) // size)
        return np.exp(2j * np.pi *
                      disp_cd *
                      self.kpt_c[:, np.newaxis])

    def new(self,
            size=None,
            pbc=None,
            kpt=None,
            comm='inherit',
            decomp=None,
            dtype=None) -> UGDesc:
        """Create new uniform grid description."""
        if decomp is None and comm == 'inherit':
            if size is None and pbc is None:
                decomp = self.decomp_cp
        comm = self.comm if comm == 'inherit' else comm
        return UGDesc(cell=self.cell_cv,
                      size=self.size_c if size is None else size,
                      pbc=self.pbc_c if pbc is None else pbc,
                      kpt=(self.kpt_c if self.kpt_c.any() else None)
                      if kpt is None else kpt,
                      comm=comm or serial_comm,
                      decomp=decomp,
                      dtype=self.dtype if dtype is None else dtype)

    def empty(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm,
              xp=np) -> UGArray:
        """Create new UGArray object.

        parameters
        ----------
        dims:
            Extra dimensions.
        comm:
            Distribute dimensions along this communicator.
        """
        return UGArray(self, dims, comm, xp=xp)

    def blocks(self, data: np.ndarray):
        """Yield views of blocks of data."""
        s0, s1, s2 = self.parsize_c
        d0_p, d1_p, d2_p = (d_p - d_p[0] for d_p in self.decomp_cp)
        for p0 in range(s0):
            b0, e0 = d0_p[p0:p0 + 2]
            for p1 in range(s1):
                b1, e1 = d1_p[p1:p1 + 2]
                for p2 in range(s2):
                    b2, e2 = d2_p[p2:p2 + 2]
                    yield data[..., b0:e0, b1:e1, b2:e2]

    def xyz(self) -> Array4D:
        """Create array of (x, y, z) coordinates."""
        indices_Rc = np.indices(self.mysize_c).transpose((1, 2, 3, 0))
        indices_Rc += self.start_c
        return indices_Rc @ (self.cell_cv.T / self.size_c).T

    def atom_centered_functions(self,
                                functions,
                                positions,
                                *,
                                atomdist=None,
                                integral=None,
                                cut=False,
                                xp=None):
        """Create UGAtomCenteredFunctions object."""
        return UGAtomCenteredFunctions(functions,
                                       positions,
                                       self,
                                       atomdist=atomdist,
                                       integral=integral,
                                       cut=cut,
                                       xp=xp)

    def transformer(self, other: UGDesc, stencil_range=3, xp=np):
        """Create transformer from one grid to another.

        (for interpolation and restriction).
        """
        from gpaw.transformers import Transformer

        apply = Transformer(self._gd, other._gd, nn=stencil_range, xp=xp).apply

        def transform(functions, out=None):
            if out is None:
                out = other.empty(functions.dims, functions.comm, xp=xp)
            for input, output in zips(functions._arrays(), out._arrays()):
                apply(input, output)
            return out

        return transform

    def eikr(self, kpt_c: Vector | None = None) -> Array3D:
        """Plane wave.

        :::
               _ _
              ik.r
             e

        Parameters
        ----------
        kpt_c:
            k-point in units of the reciprocal cell.  Defaults to the
            UGDesc objects own k-point.
        """
        if kpt_c is None:
            kpt_c = self.kpt_c
        index_Rc = np.indices(self.mysize_c).T + self.start_c
        return np.exp(2j * pi * (index_Rc @ (kpt_c / self.size_c))).T

    @property
    def _gd(self):
        # Make sure gd can be pickled (in serial):
        comm = self.comm if self.comm.size > 1 else serial_comm

        return GridDescriptor(self.size_c,
                              cell_cv=self.cell_cv,
                              pbc_c=self.pbc_c,
                              comm=comm,
                              parsize_c=[len(d_p) - 1
                                         for d_p in self.decomp_cp])

    @classmethod
    def _from_gd_and_kpt_and_dtype(cls, gd, kpt, dtype):
        return UGDesc(cell=gd.cell_cv,
                      size=gd.N_c,
                      pbc=gd.pbc_c,
                      comm=gd.comm,
                      dtype=dtype,
                      kpt=kpt,
                      decomp=gd.n_cp)

    @classmethod
    def from_cell_and_grid_spacing(cls,
                                   cell: ArrayLike1D | ArrayLike2D,
                                   grid_spacing: float,
                                   pbc=(True, True, True),
                                   kpt: Vector | None = None,
                                   comm: MPIComm = serial_comm,
                                   dtype=None) -> UGDesc:
        """Create UGDesc from grid-spacing."""
        domain = Domain(cell, pbc, kpt, comm, dtype)
        return domain.uniform_grid_with_grid_spacing(grid_spacing)

    def fft_plans(self,
                  flags: int = fftw.MEASURE,
                  xp=np,
                  dtype=None) -> fftw.FFTPlans:
        """Create FFTW-plans."""
        if dtype is None:
            dtype = self.dtype
        if self.comm.rank == 0:
            return fftw.create_plans(self.size_c, dtype, flags, xp)
        else:
            return fftw.create_plans([0, 0, 0], dtype)

    def ranks_from_fractional_positions(self,
                                        fracpos_ac: Array2D) -> Array1D:
        rank_ac = np.floor(fracpos_ac * self.parsize_c).astype(int)
        if (rank_ac < 0).any() or (rank_ac >= self.parsize_c).any():
            raise ValueError('Positions outside cell!')
        return np.ravel_multi_index(rank_ac.T, self.parsize_c)  # type: ignore

    def ecut_max(self):
        dv_cv = self.cell_cv / self.size_c[:, np.newaxis]
        return 0.5 * np.pi**2 / (dv_cv**2).sum(1).max()  # or min ??????


class UGArray(DistributedArrays[UGDesc]):
    def __init__(self,
                 grid: UGDesc,
                 dims: int | tuple[int, ...] = (),
                 comm: MPIComm = serial_comm,
                 data: np.ndarray | None = None,
                 xp=None):
        """Object for storing function(s) on a uniform grid.

        parameters
        ----------
        grid:
            Description of uniform grid.
        dims:
            Extra dimensions.
        comm:
            Distribute dimensions along this communicator.
        data:
            Data array for storage.
        """
        DistributedArrays. __init__(self, dims, grid.myshape,
                                    comm, grid.comm, data, grid.dv,
                                    grid.dtype, xp)
        self.desc = grid

    def __repr__(self):
        txt = f'UGArray(grid={self.desc}, dims={self.dims}'
        if self.comm.size > 1:
            txt += f', comm={self.comm.rank}/{self.comm.size}'
        if self.xp is not np:
            txt += ', xp=cp'
        return txt + ')'

    def new(self, data=None, zeroed=False):
        """Create new UniforGridFunctions object of same kind.

        Parameters
        ----------
        data:
            Array to use for storage.
        """
        if data is None:
            data = self.xp.empty_like(self.data)
        f_xR = UGArray(self.desc, self.dims, self.comm, data)
        if zeroed:
            f_xR.data[:] = 0.0
        return f_xR

    def __getitem__(self, index):
        data = self.data[index]
        return UGArray(data=data,
                       dims=data.shape[:-3],
                       grid=self.desc)

    def __imul__(self,
                 other: float | np.ndarray | UGArray
                 ) -> UGArray:
        if isinstance(other, float):
            self.data *= other
            return self
        if isinstance(other, UGArray):
            other = other.data
        assert other.shape[-3:] == self.data.shape[-3:]
        self.data *= other
        return self

    def __mul__(self,
                other: float | np.ndarray | UGArray
                ) -> UGArray:
        result = self.new(data=self.data.copy())
        result *= other
        return result

    def _arrays(self):
        return self.data.reshape((-1,) + self.data.shape[-3:])

    def xy(self, *axes: int | None) -> tuple[Array1D, Array1D]:
        """Extract x, y values along line.

        Useful for plotting::

          x, y = grid.xy(0, ..., 0)
          plt.plot(x, y)
        """
        assert len(axes) == 3 + len(self.dims)
        index = tuple([slice(0, None) if axis is None else axis
                       for axis in axes])
        y = self.data[index]  # type: ignore
        c = axes[-3:].index(...)
        grid = self.desc
        dx = (grid.cell_cv[c]**2).sum()**0.5 / grid.size_c[c]
        x = np.arange(grid.start_c[c], grid.end_c[c]) * dx
        return x, as_xp(y, np)

    def scatter_from(self, data=None):
        """Scatter data from rank-0 to all ranks."""
        if isinstance(data, UGArray):
            data = data.data
        comm = self.desc.comm
        if comm.size == 1:
            self.data[:] = data
            return

        if comm.rank != 0:
            comm.receive(self.data, 0, 42)
            return

        requests = []
        for rank, block in enumerate(self.desc.blocks(data)):
            if rank != 0:
                block = block.copy()
                request = comm.send(block, rank, 42, False)
                # Remember to store a reference to the
                # send buffer (block) so that is isn't
                # deallocated:
                requests.append((request, block))
            else:
                self.data[:] = block

        for request, _ in requests:
            comm.wait(request)

    def gather(self, out=None, broadcast=False):
        """Gather data from all ranks to rank-0."""
        assert out is None
        comm = self.desc.comm
        if comm.size == 1:
            return self

        if broadcast or comm.rank == 0:
            grid = self.desc.new(comm=serial_comm)
            out = grid.empty(self.dims, xp=self.xp)

        if comm.rank != 0:
            # There can be several sends before the corresponding receives
            # are posted, so use synchronous send here
            comm.ssend(self.data, 0, 301)
            if broadcast:
                comm.broadcast(out.data, 0)
                return out
            return

        # Put the subdomains from the slaves into the big array
        # for the whole domain:
        for rank, block in enumerate(self.desc.blocks(out.data)):
            if rank != 0:
                buf = self.xp.empty_like(block)
                comm.receive(buf, rank, 301)
                block[:] = buf
            else:
                block[:] = self.data

        if broadcast:
            comm.broadcast(out.data, 0)

        return out

    def fft(self, plan=None, pw=None, out=None):
        """Do FFT.

        :::

                _ _
          / _  iG.r  _
          |dr e    f(r)
          /
        """
        assert self.dims == ()
        if out is None:
            assert pw is not None
            out = pw.empty(xp=self.xp)
        if pw is None:
            pw = out.desc
        input = self
        if self.desc.comm.size > 1:
            input = input.gather()
        if self.desc.comm.rank == 0:
            plan = plan or self.desc.fft_plans(xp=self.xp)
            coefs = plan.fft_sphere(input.data, pw)
        else:
            coefs = None

        out.scatter_from(coefs)

        return out

    def norm2(self):
        """Calculate integral over cell of absolute value squared.

        :::

         /   _  2 _
         ||a(r)| dr
         /
        """
        norm_x = []
        arrays_xR = self._arrays()
        for a_R in arrays_xR:
            norm_x.append(np.vdot(a_R, a_R).real * self.desc.dv)
        result = np.array(norm_x).reshape(self.mydims)
        self.desc.comm.sum(result)
        return result

    def integrate(self, other=None, skip_sum=False):
        """Integral of self or self times cc(other)."""
        if other is not None:
            assert self.desc.dtype == other.desc.dtype
            a_xR = self._arrays()
            b_yR = other._arrays()
            a_xR = a_xR.reshape((len(a_xR), -1))
            b_yR = b_yR.reshape((len(b_yR), -1))
            result = (a_xR @ b_yR.T.conj()).reshape(self.dims + other.dims)
        else:
            # Make sure we have an array and not a scalar!
            result = self.xp.asarray(self.data.sum(axis=(-3, -2, -1)))

        if not skip_sum:
            self.desc.comm.sum(result)
        if result.ndim == 0:
            result = result.item()  # convert to scalar
        return result * self.desc.dv

    def to_pbc_grid(self):
        """Convert to UniformGrud with ``pbc=(True, True, True)``."""
        if self.desc.pbc_c.all():
            return self
        grid = self.desc.new(pbc=True)
        new = grid.empty(self.dims)
        new.data[:] = 0.0
        *_, i, j, k = self.data.shape
        new.data[..., -i:, -j:, -k:] = self.data
        return new

    def multiply_by_eikr(self, kpt_c: Vector | None = None) -> None:
        """Multiply by `exp(ik.r)`."""
        if kpt_c is None:
            kpt_c = self.desc.kpt_c
        else:
            kpt_c = np.asarray(kpt_c)
        if kpt_c.any():
            self.data *= self.desc.eikr(kpt_c)

    def interpolate(self,
                    plan1: fftw.FFTPlans | None = None,
                    plan2: fftw.FFTPlans | None = None,
                    grid: UGDesc | None = None,
                    out: UGArray | None = None) -> UGArray:
        """Interpolate to finer grid.

        Parameters
        ----------
        plan1:
            Plan for FFT (course grid).
        plan2:
            Plan for inverse FFT (fine grid).
        grid:
            Target grid.
        out:
            Target UGArray object.
        """
        if out is None:
            if grid is None:
                raise ValueError('Please specify "grid" or "out".')
            out = grid.empty(self.dims, xp=self.xp)

        if not out.desc.pbc_c.all() or not self.desc.pbc_c.all():
            raise ValueError('Grids must have pbc=True!')

        if self.desc.comm.size > 1:
            input = self.gather()
            if input:
                output = input.interpolate(plan1, plan2,
                                           out.desc.new(comm=None))
                out.scatter_from(output.data)
            else:
                out.scatter_from()
            return out

        size1_c = self.desc.size_c
        size2_c = out.desc.size_c
        if (size2_c <= size1_c).any():
            raise ValueError('Too few points in target grid!')

        plan1 = plan1 or self.desc.fft_plans(xp=self.xp)
        plan2 = plan2 or out.desc.fft_plans(xp=self.xp)

        if self.dims:
            for input, output in zips(self.flat(), out.flat()):
                input.interpolate(plan1, plan2, grid, output)
            return out

        plan1.tmp_R[:] = self.data
        kpt_c = self.desc.kpt_c
        if kpt_c.any():
            plan1.tmp_R *= self.desc.eikr(-kpt_c)
        plan1.fft()

        a_Q = plan1.tmp_Q
        b_Q = plan2.tmp_Q

        e0, e1, e2 = 1 - size1_c % 2  # even or odd size
        a0, a1, a2 = size2_c // 2 - size1_c // 2
        b0, b1, b2 = size1_c + (a0, a1, a2)

        if self.desc.dtype == float:
            b2 = (b2 - a2) // 2 + 1
            a2 = 0
            axes = [0, 1]
        else:
            axes = [0, 1, 2]

        b_Q[:] = 0.0
        b_Q[a0:b0, a1:b1, a2:b2] = self.xp.fft.fftshift(a_Q, axes=axes)

        if e0:
            b_Q[a0, a1:b1, a2:b2] *= 0.5
            b_Q[b0, a1:b1, a2:b2] = b_Q[a0, a1:b1, a2:b2]
            b0 += 1
        if e1:
            b_Q[a0:b0, a1, a2:b2] *= 0.5
            b_Q[a0:b0, b1, a2:b2] = b_Q[a0:b0, a1, a2:b2]
            b1 += 1
        if self.desc.dtype == complex:
            if e2:
                b_Q[a0:b0, a1:b1, a2] *= 0.5
                b_Q[a0:b0, a1:b1, b2] = b_Q[a0:b0, a1:b1, a2]
        else:
            if e2:
                b_Q[a0:b0, a1:b1, b2 - 1] *= 0.5

        b_Q[:] = self.xp.fft.ifftshift(b_Q, axes=axes)
        plan2.ifft()
        out.data[:] = plan2.tmp_R
        out.data *= (1.0 / self.data.size)
        out.multiply_by_eikr()
        return out

    def fft_restrict(self,
                     plan1: fftw.FFTPlans | None = None,
                     plan2: fftw.FFTPlans | None = None,
                     grid: UGDesc | None = None,
                     out: UGArray | None = None) -> UGArray:
        """Restrict to coarser grid.

        Parameters
        ----------
        plan1:
            Plan for FFT.
        plan2:
            Plan for inverse FFT.
        grid:
            Target grid.
        out:
            Target UGArray object.
        """
        if out is None:
            if grid is None:
                raise ValueError('Please specify "grid" or "out".')
            out = grid.empty(self.dims, xp=self.xp)

        if not out.desc.pbc_c.all() or not self.desc.pbc_c.all():
            raise ValueError('Grids must have pbc=True!')

        if self.desc.comm.size > 1:
            input = self.gather()
            if input:
                output = input.fft_restrict(plan1, plan2,
                                            out.desc.new(comm=None))
                out.scatter_from(output.data)
            else:
                out.scatter_from()
            return out

        size1_c = self.desc.size_c
        size2_c = out.desc.size_c

        plan1 = plan1 or self.desc.fft_plans()
        plan2 = plan2 or out.desc.fft_plans()

        if self.dims:
            for input, output in zips(self.flat(), out.flat()):
                input.fft_restrict(plan1, plan2, grid, output)
            return out

        plan1.tmp_R[:] = self.data
        a_Q = plan2.tmp_Q
        b_Q = plan1.tmp_Q

        e0, e1, e2 = 1 - size2_c % 2  # even or odd size
        a0, a1, a2 = size1_c // 2 - size2_c // 2
        b0, b1, b2 = size2_c // 2 + size1_c // 2 + 1

        if self.desc.dtype == float:
            b2 = size2_c[2] // 2 + 1
            a2 = 0
            axes = [0, 1]
        else:
            axes = [0, 1, 2]

        plan1.fft()
        b_Q[:] = self.xp.fft.fftshift(b_Q, axes=axes)

        if e0:
            b_Q[a0, a1:b1, a2:b2] += b_Q[b0 - 1, a1:b1, a2:b2]
            b_Q[a0, a1:b1, a2:b2] *= 0.5
            b0 -= 1
        if e1:
            b_Q[a0:b0, a1, a2:b2] += b_Q[a0:b0, b1 - 1, a2:b2]
            b_Q[a0:b0, a1, a2:b2] *= 0.5
            b1 -= 1
        if self.desc.dtype == complex and e2:
            b_Q[a0:b0, a1:b1, a2] += b_Q[a0:b0, a1:b1, b2 - 1]
            b_Q[a0:b0, a1:b1, a2] *= 0.5
            b2 -= 1

        a_Q[:] = b_Q[a0:b0, a1:b1, a2:b2]
        a_Q[:] = self.xp.fft.ifftshift(a_Q, axes=axes)
        plan2.ifft()
        out.data[:] = plan2.tmp_R
        out.data *= (1.0 / self.data.size)
        return out

    def abs_square(self,
                   weights: Array1D,
                   out: UGArray | None = None) -> None:
        """Add weighted absolute square of data to output array."""
        assert out is not None
        for f, psit_R in zips(weights, self.data):
            add_to_density(f, psit_R, out.data)

    def symmetrize(self, rotation_scc, translation_sc):
        """Make data symmetric."""
        if len(rotation_scc) == 1:
            return

        a_xR = self.gather()

        if a_xR is None:
            b_xR = None
        else:
            if self.xp is not np:
                a_xR = a_xR.to_xp(np)
            b_xR = a_xR.new()
            t_sc = (translation_sc * self.desc.size_c).round().astype(int)
            offset_c = 1 - self.desc.pbc_c
            for a_R, b_R in zips(a_xR._arrays(), b_xR._arrays()):
                b_R[:] = 0.0
                for r_cc, t_c in zips(rotation_scc, t_sc):
                    symmetrize_ft(a_R, b_R, r_cc, t_c, offset_c)
            if self.xp is not np:
                b_xR = b_xR.to_xp(self.xp)
        self.scatter_from(b_xR)

        self.data *= 1.0 / len(rotation_scc)

    def randomize(self, seed: int | None = None) -> None:
        """Insert random numbers between -0.5 and 0.5 into data."""
        if seed is None:
            seed = self.comm.rank + self.desc.comm.rank * self.comm.size
        rng = np.random.default_rng(seed)
        a = self.data.view(float)
        rng.random(a.shape, out=a)
        a -= 0.5

    def moment(self, center_v=None):
        """Calculate moment of data."""
        assert self.dims == ()
        ug = self.desc

        index_cr = [np.arange(ug.start_c[c], ug.end_c[c], dtype=float)
                    for c in range(3)]

        if center_v is not None:
            frac_c = np.linalg.solve(ug.cell_cv.T, center_v)
            corner_c = (frac_c % 1 - 0.5) * ug.size_c
            for index_r, corner, size in zips(index_cr, corner_c, ug.size_c):
                index_r -= corner
                index_r %= size
                index_r += corner

        rho_ijk = self.data
        rho_ij = rho_ijk.sum(axis=2)
        rho_ik = rho_ijk.sum(axis=1)
        rho_cr = [rho_ij.sum(axis=1), rho_ij.sum(axis=0), rho_ik.sum(axis=0)]
        if self.xp is not np:
            rho_cr = [rho_r.get() for rho_r in rho_cr]

        d_c = [index_r @ rho_r for index_r, rho_r in zips(index_cr, rho_cr)]
        d_v = (d_c / ug.size_c) @ ug.cell_cv * self.dv
        self.desc.comm.sum(d_v)
        return d_v

    def scaled(self, s: float, v: float = 1.0):
        """Create new scaled UGArray object.

        Unit cell axes are multiplied by `s` and data by `v`.
        """
        grid = self.desc
        grid = UGDesc(cell=grid.cell_cv * s,
                      size=grid.size_c,
                      pbc=grid.pbc_c,
                      kpt=(grid.kpt_c if grid.kpt_c.any() else None),
                      dtype=grid.dtype,
                      comm=grid.comm)
        return UGArray(grid, self.dims, self.comm, self.data * v)

    def add_ked(self,
                occ_n: Array1D,
                taut_R: UGArray) -> None:
        grad_v = [
            Gradient(self.desc._gd, v, n=3, dtype=self.desc.dtype)
            for v in range(3)]
        tmp_R = self.desc.empty()
        for f, psit_R in zips(occ_n, self):
            for grad in grad_v:
                grad(psit_R, tmp_R)
                add_to_density(0.5 * f, tmp_R.data, taut_R.data)

    def redist(self,
               domain: UGDesc,
               comm1: MPIComm, comm2: MPIComm) -> UGArray:
        a = super().redist(domain, comm1, comm2)
        assert isinstance(a, UGArray)
        return a



from ase import Atom

from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.analyse.eed import ExteriorElectronDensity


def test_utilities_eed(in_tmp_dir):
    fwfname = 'H2_kpt441_wf.gpw'

    # write first if needed
    s = Cluster([Atom('H'), Atom('H', [0, 0, 1])], pbc=[1, 1, 0])
    s.minimal_box(3.)
    c = GPAW(mode='fd', xc='PBE', h=.3, kpts=(4, 4, 1),
             convergence={'density': 1e-3, 'eigenstates': 1e-3})
    c.calculate(s)
    c.write(fwfname, 'all')

    eed = ExteriorElectronDensity(c.wfs.gd, s)
    eed.write_mies_weights(c.wfs)



from __future__ import annotations
from pathlib import Path
from typing import Union, List, Optional, Sequence, TYPE_CHECKING

import numpy as np
from ase.dft.dos import linear_tetrahedron_integration as lti

from gpaw.setup import Setup
from gpaw.spinorbit import soc_eigenstates, BZWaveFunctions
from gpaw.typing import Array1D, Array2D, Array3D, ArrayLike1D

if TYPE_CHECKING:
    from gpaw.calculator import GPAW
    from gpaw.new.ase_interface import ASECalculator


class IBZWaveFunctions:
    """Container for eigenvalues and PAW projections (only IBZ)."""
    def __init__(self, calc: ASECalculator | GPAW):
        self.calc = calc
        self.fermi_level = self.calc.get_fermi_level()
        self.size = calc.wfs.kd.N_c
        self.bz2ibz_map = calc.wfs.kd.bz2ibz_k

    def weights(self) -> Array1D:
        """Weigths of IBZ k-points (adds to 1.0)."""
        return self.calc.wfs.kd.weight_k

    def eigenvalues(self) -> Array3D:
        """All eigenvalues."""
        kd = self.calc.wfs.kd
        eigs = np.array([[self.calc.get_eigenvalues(kpt=k, spin=s)
                          for k in range(kd.nibzkpts)]
                         for s in range(kd.nspins)])
        return eigs

    def pdos_weights(self,
                     a: int,
                     indices: List[int]
                     ) -> Array3D:
        """Projections for PDOS.

        Returns (nibzkpts, nbands, nspins)-shaped ndarray
        of the square of absolute value of the projections.  The *indices*
        list contains projector-indices.
        """
        kd = self.calc.wfs.kd
        dos_kns = np.zeros((kd.nibzkpts,
                            self.calc.wfs.bd.nbands,
                            kd.nspins))
        bands = self.calc.wfs.bd.get_slice()

        for wf in self.calc.wfs.kpt_u:
            P_ani = wf.projections
            if a in P_ani:
                P_ni = P_ani[a][:, indices]
                dos_kns[wf.k, bands, wf.s] = (abs(P_ni)**2).sum(1)

        self.calc.world.sum(dos_kns)
        return dos_kns


def get_projector_numbers(setup: Setup, ell: int) -> List[int]:
    """Find indices of bound-state PAW projector functions.

    >>> from gpaw.setup import create_setup
    >>> setup = create_setup('Li')
    >>> get_projector_numbers(setup, 0)
    [0]
    >>> get_projector_numbers(setup, 1)
    [1, 2, 3]
    """
    indices = []
    i1 = 0
    for n, l in zip(setup.n_j, setup.l_j):
        i2 = i1 + 2 * l + 1
        if l == ell and n >= 0:
            indices += list(range(i1, i2))
        i1 = i2
    return indices


def gaussian_dos(eig_kn,
                 weight_kn,
                 weight_k,
                 energies,
                 width: float) -> Array1D:
    """Simple broadening with a Gaussian."""
    dos = np.zeros_like(energies)
    if weight_kn is None:
        for e_n, w in zip(eig_kn, weight_k):
            for e in e_n:
                dos += w * np.exp(-((energies - e) / width)**2)
    else:
        for e_n, w, w_n in zip(eig_kn, weight_k, weight_kn):
            for e, w2 in zip(e_n, w_n):
                dos += w * w2 * np.exp(-((energies - e) / width)**2)
    return dos / (np.pi**0.5 * width)


def linear_tetrahedron_dos(eig_kn,
                           weight_kn,
                           energies,
                           cell,
                           size,
                           bz2ibz_map=None) -> Array1D:
    """Linear-tetrahedron method."""
    if len(eig_kn) != np.prod(size):
        eig_kn = eig_kn[bz2ibz_map]
        if weight_kn is not None:
            weight_kn = weight_kn[bz2ibz_map]

    shape = tuple(size) + (-1,)
    eig_kn = eig_kn.reshape(shape)
    if weight_kn is not None:
        weight_kn = weight_kn.reshape(shape)

    dos = lti(cell, eig_kn, energies, weight_kn)
    return dos


class DOSCalculator:
    def __init__(self,
                 wfs,
                 setups=None,
                 cell=None,
                 shift_fermi_level=True):
        self.wfs = wfs
        self.setups = setups
        self.cell = cell

        self.eig_skn = wfs.eigenvalues()
        self.fermi_level = wfs.fermi_level

        if shift_fermi_level:
            self.eig_skn -= wfs.fermi_level

        self.collinear = (self.eig_skn.ndim == 3)
        if self.collinear:
            self.degeneracy = 2 / len(self.eig_skn)
        else:
            self.eig_skn = np.array([self.eig_skn, self.eig_skn])
            self.degeneracy = 0.5

        self.nspins = len(self.eig_skn)
        self.weight_k = wfs.weights()

    def get_energies(self,
                     emin: Optional[float] = None,
                     emax: Optional[float] = None,
                     npoints: int = 100):
        emin = emin if emin is not None else self.eig_skn.min()
        emax = emax if emax is not None else self.eig_skn.max()
        return np.linspace(emin, emax, npoints)

    @classmethod
    def from_calculator(cls,
                        filename: ASECalculator | GPAW | Path | str,
                        soc=False, theta=0.0, phi=0.0,
                        shift_fermi_level=True):
        """Create DOSCalculator from a GPAW calculation.

        filename: str
            Name of restart-file or GPAW calculator object.
        """
        from gpaw.calculator import GPAW
        if not isinstance(filename, (str, Path)):
            calc = filename
        else:
            calc = GPAW(filename, txt=None)

        wfs: Union[BZWaveFunctions, IBZWaveFunctions]
        if soc:
            wfs = soc_eigenstates(calc, theta=theta, phi=phi)
        else:
            wfs = IBZWaveFunctions(calc)

        return DOSCalculator(wfs,
                             calc.setups, calc.atoms.cell,
                             shift_fermi_level)

    def calculate(self,
                  energies: ArrayLike1D,
                  eig_kn: Array2D,
                  weight_kn: Array2D = None,
                  width: float = 0.1):
        energies = np.asarray(energies)
        if width > 0.0:
            return gaussian_dos(eig_kn, weight_kn,
                                self.weight_k, energies, width)
        else:
            return linear_tetrahedron_dos(
                eig_kn, weight_kn, energies,
                self.cell, self.wfs.size, self.wfs.bz2ibz_map)

    def raw_dos(self,
                energies: Sequence[float],
                spin: Union[int, None] = None,
                width: float = 0.1) -> Array1D:
        """Calculate density of states.

        width: float
            Width of Gaussians in eV.  Use width=0.0 to use the
            linear-tetrahedron-interpolation method.
        """
        if spin is None:
            dos = sum(self.calculate(energies, eig_kn, width=width)
                      for eig_kn in self.eig_skn)
            dos *= self.degeneracy
        else:
            dos = self.calculate(energies, self.eig_skn[spin], width=width)

        return dos

    def raw_pdos(self,
                 energies: Sequence[float],
                 a: int,
                 l: int,
                 m: Optional[int] = None,
                 spin: int = None,
                 width: float = 0.1) -> Array1D:
        """Calculate projected density of states.

        a:
            Atom index.
        l:
            Angular momentum quantum number.
        m:
            Magnetic quantum number.  Default is None meaning sum over all m.
            For p-orbitals, m=0,1,2 translates to y, z and x.
            For d-orbitals, m=0,1,2,3,4 translates to xy, yz, 3z2-r2,
            zx and x2-y2.
        spin:
            Must be 0, 1 or None meaning spin-up, down or total respectively.
        width: float
            Width of Gaussians in eV.  Use width=0.0 to use the
            linear-tetrahedron-interpolation method.
        """
        indices = get_projector_numbers(self.setups[a], l)
        if m is not None:
            indices = indices[m::(2 * l) + 1]
        weight_kns = self.wfs.pdos_weights(a, indices)

        if spin is None:
            dos = sum(self.calculate(energies,
                                     eig_kn,
                                     weight_nk.T,
                                     width=width)
                      for eig_kn, weight_nk
                      in zip(self.eig_skn, weight_kns.T))
            dos *= self.degeneracy
        else:
            dos = self.calculate(energies,
                                 self.eig_skn[spin],
                                 weight_kns[:, :, spin],
                                 width=width)

        return dos



from math import pi, sqrt
import numpy as np
from numpy import exp
from scipy.special import erf, erfc

from gpaw.poisson import FDPoissonSolver
from gpaw.utilities.gauss import Gaussian
from gpaw.fd_operators import FDOperator, laplace
from gpaw.transformers import Transformer


class HelmholtzGaussian(Gaussian):
    def get_phi(self, k2):
        """Get the solution of the Helmholtz equation for a Gaussian."""
#   This should lead to very big errors
        r = np.sqrt(self.r2)
        k = sqrt(k2)
        sigma = 1. / sqrt(2 * self.a)
        p = sigma * k / sqrt(2)
        i = 1j
        rhop = r / sqrt(2) / sigma + i * p
        rhom = r / sqrt(2) / sigma - i * p
        # h = np.sin(k * r)

        # the gpaw-Gaussian is sqrt(4 * pi) times a 3D normalized Gaussian
        return sqrt(4 * pi) * exp(-p**2) / r / 2 * (
            np.cos(k * r) * (erf(rhop) + erf(rhom)) +
            i * np.sin(k * r) * (2 + erf(rhop) - erf(rhom)))


class ScreenedPoissonGaussian(Gaussian):
    def get_phi(self, mu2):
        """Get the solution of the screened Poisson equation for a Gaussian.

           The Gaussian is centered to middle of grid-descriptor."""
        r = np.sqrt(self.r2)
        mu = sqrt(mu2)
        sigma = 1. / sqrt(2 * self.a)
        sig2 = sigma**2
        mrho = (sig2 * mu - r) / (sqrt(2) * sigma)
        prho = (sig2 * mu + r) / (sqrt(2) * sigma)

        # the gpaw-Gaussian is sqrt(4 * pi) times a 3D normalized Gaussian
        return sqrt(4 * pi) * exp(sig2 * mu2 / 2.0) / (2 * r) * (
            exp(-mu * r) * erfc(mrho) - exp(mu * r) * erfc(prho))


class HelmholtzOperator(FDOperator):
    def __init__(self, gd, scale=1.0, n=1, dtype=float, k2=0.0):
        """Helmholtz for general non orthorhombic grid.

        gd: GridDescriptor
            Descriptor for grid.
        scale: float
            Scaling factor.  Use scale=-0.5 for a kinetic energy operator.
        n: int
            Range of stencil.  Stencil has O(h^(2n)) error.
        dtype: float or complex
            Data-type to work on.
        """

        # Order the 13 neighbor grid points:
        M_ic = np.indices((3, 3, 3)).reshape((3, -3)).T[-13:] - 1
        u_cv = gd.h_cv / (gd.h_cv**2).sum(1)[:, np.newaxis]**0.5
        u2_i = (np.dot(M_ic, u_cv)**2).sum(1)
        i_d = u2_i.argsort()

        m_mv = np.array([(2, 0, 0), (0, 2, 0), (0, 0, 2),
                         (0, 1, 1), (1, 0, 1), (1, 1, 0)])
        # Try 3, 4, 5 and 6 directions:
        for D in range(3, 7):
            h_dv = np.dot(M_ic[i_d[:D]], gd.h_cv)
            A_md = (h_dv**m_mv[:, np.newaxis, :]).prod(2)
            a_d, residual, rank, s = np.linalg.lstsq(A_md, [1, 1, 1, 0, 0, 0],
                                                     rcond=-1)
            if residual.sum() < 1e-14:
                assert rank == D, 'You have a weird unit cell!'
                # D directions was OK
                break

        a_d *= scale
        offsets = [(0, 0, 0)]
        coefs = [laplace[n][0] * a_d.sum()]
        coefs[0] += k2 * scale
        for d in range(D):
            M_c = M_ic[i_d[d]]
            offsets.extend(np.arange(1, n + 1)[:, np.newaxis] * M_c)
            coefs.extend(a_d[d] * np.array(laplace[n][1:]))
            offsets.extend(np.arange(-1, -n - 1, -1)[:, np.newaxis] * M_c)
            coefs.extend(a_d[d] * np.array(laplace[n][1:]))

        FDOperator.__init__(self, coefs, offsets, gd, dtype)

        self.description = (
            '%d*%d+1=%d point O(h^%d) finite-difference Helmholtz' %
            ((self.npoints - 1) // n, n, self.npoints, 2 * n))


class HelmholtzSolver(FDPoissonSolver):
    """Solve the Helmholtz or screened Poisson equations.

       The difference between the Helmholtz equation:

           (Laplace + k^2) phi = n

       and the screened Poisson equation:

           (Laplace - mu^2) phi = n

       is only the sign of the added inhomogenity. Because of
       this we can use one class to solve both. So if k2 is
       greater zero we'll try to solve the Helmhlotz equation,
       otherwise we'll try to solve the screened Poisson equation.
    """

    def __init__(self, k2=0.0, nn='M', relax='GS', eps=2e-10,
                 use_charge_center=True):
        assert k2 <= 0, 'Currently only defined for k^2<=0'
        FDPoissonSolver.__init__(self, nn, relax, eps,
                                 use_charge_center=use_charge_center)
        self.k2 = k2

    def set_grid_descriptor(self, gd):
        # Should probably be renamed initialize
        self.gd = gd
        self.dv = gd.dv

        gd = self.gd
        scale = -0.25 / pi

        if self.nn == 'M':
            raise ValueError(
                'Helmholtz not defined for Mehrstellen stencil')
        self.operators = [HelmholtzOperator(gd, scale, self.nn, k2=self.k2)]
        self.B = None

        self.interpolators = []
        self.restrictors = []

        level = 0
        self.presmooths = [2]
        self.postsmooths = [1]

        # Weights for the relaxation,
        # only used if 'J' (Jacobi) is chosen as method
        self.weights = [2.0 / 3.0]

        while level < 4:
            try:
                gd2 = gd.coarsen()
            except ValueError:
                break
            self.operators.append(HelmholtzOperator(gd2, scale, 1,
                                                    k2=self.k2))
            self.interpolators.append(Transformer(gd2, gd))
            self.restrictors.append(Transformer(gd, gd2))
            self.presmooths.append(4)
            self.postsmooths.append(4)
            self.weights.append(1.0)
            level += 1
            gd = gd2

        self.levels = level

        if self.relax_method == 1:
            self.description = 'Gauss-Seidel'
        else:
            self.description = 'Jacobi'
        self.description += ' solver with %d multi-grid levels' % (level + 1)
        self.description += '\nStencil: ' + self.operators[0].description

    def load_gauss(self, center=None):
        """Load the gaussians."""
        if not hasattr(self, 'rho_gauss') or center is not None:
            if self.k2 > 0:
                gauss = HelmholtzGaussian(self.gd, center=center)
            else:
                gauss = ScreenedPoissonGaussian(self.gd, center=center)
            self.rho_gauss = gauss.get_gauss(0)
            self.phi_gauss = gauss.get_phi(abs(self.k2))



import numpy as np
import pytest
from ase.build import molecule

from gpaw import GPAW, Mixer
from gpaw.mpi import world

# Place one atom next to cell boundary, then check that reuse_wfs
# works correctly when atom is subsequently displaced across the
# boundary, i.e., that the kpoint phases of the PAW correction
# are handled correctly when unprojecting/reprojecting the wavefunctions.


@pytest.mark.later
def test_reuse_wfs_celldisp(in_tmp_dir):
    def check(reuse):
        atoms = molecule('H2')
        atoms.pbc = 1
        atoms.center(vacuum=1.5)
        atoms.positions -= atoms.positions[1]
        dz = 1e-2
        atoms.positions[:, 2] += dz

        calc = GPAW(mode='pw',
                    txt='gpaw.txt',
                    nbands=1,
                    experimental=dict(
                        reuse_wfs_method='paw' if reuse else None),
                    kpts=[[-0.3, 0.4, 0.2]],
                    symmetry='off',
                    mixer=Mixer(0.7, 5, 50.0))
        atoms.calc = calc

        first_iter_err = []

        def monitor():
            logerr = np.log10(calc.wfs.eigensolver.error)
            n = calc.scf.niter
            if n == 1:
                first_iter_err.append(logerr)

            if world.rank == 0:
                print('iter', n, 'err', logerr)
        calc.attach(monitor, 1)

        atoms.get_potential_energy()
        logerr1 = first_iter_err.pop()

        atoms.positions[:, 2] -= 2 * dz
        atoms.get_potential_energy()

        logerr2 = first_iter_err.pop()

        if world.rank == 0:
            print('reuse={}'.format(bool(reuse)))
            print('logerr1', logerr1)
            print('logerr2', logerr2)
            gain = logerr2 - logerr1
            print('gain', gain)
        return logerr2

    noreuse_logerr = check(0)
    reuse_logerr = check(1)
    # Ref values: logerr=-3.6 without reuse_wfs and -5.0 with reuse_wfs
    assert reuse_logerr < -4.8, reuse_logerr
    assert reuse_logerr < noreuse_logerr - 1.2, (reuse_logerr, noreuse_logerr)



import numpy as np

from gpaw.mpi import world
from gpaw.typing import ArrayND


# NOTE: This routine is not specific to Raman per se. Maybe it should go
#       somewhere else?
def get_dipole_transitions(wfs) -> ArrayND:
    r"""
    Finds dipole transition matrix elements based on the velocity form.

    Dipole and momentum matrix elements are related by the expression:
    <nk|r|mk> = -i hbar/m <nk|p|mk> / (E_nk-E_mk)

    For n=m this is ill defined, and element will be set to zero.

    Parameters
    ----------
    wfs
        LCAO WaveFunctions object
    """
    p_skvnm = get_momentum_transitions(wfs, False)
    r_skvnm = np.zeros_like(p_skvnm)
    for kpt in wfs.kpt_u:
        # NOTE: Check whether it's this way or other way around.
        deltaE = kpt.eps_n[:, None] - kpt.eps_n[None, :]
        np.fill_diagonal(deltaE, np.inf)
        r_skvnm[kpt.s, kpt.k, :] = -1j * p_skvnm[kpt.s, kpt.k, :] / deltaE
    wfs.kd.comm.sum(r_skvnm)
    return r_skvnm


def get_momentum_transitions(wfs, savetofile: bool = True) -> ArrayND:
    r"""
    Finds the momentum matrix elements:
    <nk|p|mk> = k \delta_nm - i <nk|\nabla|mk>

    Parameters
    ----------
    wfs
        LCAO WaveFunctions object
    savetofile: bool
        Determines whether matrix is written to the
        file mom_skvnm.npy (default=True)
    """
    assert wfs.bd.comm.size == 1
    assert wfs.mode == 'lcao'
    nbands = wfs.bd.nbands
    nspins = wfs.nspins
    nk = wfs.kd.nibzkpts
    gd = wfs.gd
    dtype = wfs.dtype
    ksl = wfs.ksl

    # print(wfs.kd.comm.size, wfs.gd.comm.size, wfs.bd.comm.size)

    mom_skvnm = np.zeros((nspins, nk, 3, nbands, nbands), dtype=complex)

    dThetadR_qvMM, dTdR_qvMM = wfs.manytci.O_qMM_T_qMM(gd.comm, ksl.Mstart,
                                                       ksl.Mstop, False,
                                                       derivative=True)

    mome_skvnm = np.zeros((nspins, nk, 3, nbands, nbands), dtype=dtype)
    momd_skvnm = np.zeros((nspins, nk, 3, nbands, nbands), dtype=dtype)
    moma_skvnm = np.zeros((nspins, nk, 3, nbands, nbands), dtype=dtype)

    B_cv = 2.0 * np.pi * gd.icell_cv
    for kpt in wfs.kpt_u:
        C_nM = kpt.C_nM
        for v in range(3):
            dThetadRv_MM = dThetadR_qvMM[kpt.q, v]
            nabla_nn = -(C_nM.conj() @ dThetadRv_MM.conj() @ C_nM.T)
            gd.comm.sum(nabla_nn)
            mome_skvnm[kpt.s, kpt.k, v] = nabla_nn

        # augmentation part
        moma_vnm = np.zeros((3, nbands, nbands), dtype=dtype)
        for a, P_ni in kpt.P_ani.items():
            nabla_iiv = wfs.setups[a].nabla_iiv
            moma_vnm += np.einsum('ni,ijv,mj->vnm',
                                  P_ni.conj(), nabla_iiv, P_ni)
        gd.comm.sum(moma_vnm)
        moma_skvnm[kpt.s, kpt.k] = moma_vnm

        # diagonal term
        k_v = np.dot(wfs.kd.ibzk_kc[kpt.k], B_cv)
        momd_skvnm[kpt.s, kpt.k] = np.einsum("k,nm->knm",
                                             k_v, np.identity(nbands))

    mom_skvnm = momd_skvnm - 1j * (mome_skvnm + moma_skvnm)
    wfs.kd.comm.sum(mom_skvnm)

    if world.rank == 0 and savetofile:
        np.save('mom_skvnm.npy', mom_skvnm)
    return mom_skvnm



import json
import re
from typing import Tuple

import numpy as np
from ase import Atoms
from ase.units import Bohr
from ase.utils import IOContext
from gpaw.fd_operators import Gradient
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.observer import TDDFTObserver
from gpaw.typing import Vector
from gpaw.utilities.tools import coordinates


def calculate_magnetic_moment_on_grid(wfs, grad_v, r_vG, dM_vaii, *,
                                      only_pseudo=False):
    """Calculate magnetic moment on grid.

    Parameters
    ----------
    wfs
        Wave functions object
    grad_v
        List of gradient operators
    r_vG
        Grid point coordinates
    dM_vaii
        Atomic PAW corrections for magnetic moment
    only_pseudo
        If true, do not add atomic corrections

    Returns
    -------
    Magnetic moment vector
    """
    gd = wfs.gd
    mode = wfs.mode
    bd = wfs.bd
    kpt_u = wfs.kpt_u

    rxnabla_v = np.zeros(3, dtype=complex)
    if mode == 'lcao':
        psit_G = gd.empty(dtype=complex)
    nabla_psit_vG = gd.empty(3, dtype=complex)
    for kpt in kpt_u:
        for n, f in enumerate(kpt.f_n):
            if mode == 'lcao':
                psit_G[:] = 0.0
                wfs.basis_functions.lcao_to_grid(kpt.C_nM[n], psit_G, kpt.q)
            else:
                psit_G = kpt.psit_nG[n]

            for v in range(3):
                grad_v[v].apply(psit_G, nabla_psit_vG[v], kpt.phase_cd)

            # rxnabla   = <psi1| r x nabla |psi2>
            # rxnabla_x = <psi1| r_y nabla_z - r_z nabla_y |psi2>
            # rxnabla_y = <psi1| r_z nabla_x - r_x nabla_z |psi2>
            # rxnabla_z = <psi1| r_x nabla_y - r_y nabla_x |psi2>
            for v in range(3):
                v1 = (v + 1) % 3
                v2 = (v + 2) % 3
                rnabla_psit_G = (r_vG[v1] * nabla_psit_vG[v2]
                                 - r_vG[v2] * nabla_psit_vG[v1])
                rxnabla_v[v] += f * gd.integrate(psit_G.conj() * rnabla_psit_G)

    if not only_pseudo:
        paw_rxnabla_v = np.zeros(3, dtype=complex)
        for kpt in kpt_u:
            for v in range(3):
                for a, P_ni in kpt.P_ani.items():
                    paw_rxnabla_v[v] += np.einsum('n,ni,ij,nj',
                                                  kpt.f_n, P_ni.conj(),
                                                  dM_vaii[v][a], P_ni,
                                                  optimize=True)
        gd.comm.sum(paw_rxnabla_v)
        rxnabla_v += paw_rxnabla_v

    bd.comm.sum(rxnabla_v)
    return -0.5 * rxnabla_v.imag


def calculate_magnetic_moment_atomic_corrections(R_av, setups, partition):
    """Calculate atomic PAW augmentation corrections for magnetic moment.

    Parameters
    ----------
    R_av
        Atom positions
    setups
        PAW setups object
    partition
        Atom partition object

    Returns
    -------
    Atomic correction matrices
    """
    # augmentation contributions to magnetic moment
    # <psi1| r x nabla |psi2> = <psi1| (r - Ra + Ra) x nabla |psi2>
    #                         = <psi1| (r - Ra) x nabla |psi2>
    #                             + Ra x <psi1| nabla |psi2>

    def shape(a):
        ni = setups[a].ni
        return ni, ni

    dM_vaii = []
    for _ in range(3):
        dM_aii = partition.arraydict(shapes=shape, dtype=complex)
        dM_vaii.append(dM_aii)

    for a in partition.my_indices:
        Ra_v = R_av[a]
        rxnabla_iiv = setups[a].rxnabla_iiv
        nabla_iiv = setups[a].nabla_iiv

        # rxnabla   = <psi1| (r - Ra) x nabla |psi2>
        # Rxnabla   = Ra x <psi1| nabla |psi2>
        # Rxnabla_x = Ra_y nabla_z - Ra_z nabla_y
        # Rxnabla_y = Ra_z nabla_x - Ra_x nabla_z
        # Rxnabla_z = Ra_x nabla_y - Ra_y nabla_x
        for v in range(3):
            v1 = (v + 1) % 3
            v2 = (v + 2) % 3
            Rxnabla_ii = (Ra_v[v1] * nabla_iiv[:, :, v2]
                          - Ra_v[v2] * nabla_iiv[:, :, v1])
            dM_vaii[v][a][:] = Rxnabla_ii + rxnabla_iiv[:, :, v]

    return dM_vaii


def calculate_magnetic_moment_matrix(kpt_u, bfs, correction, r_vG, dM_vaii, *,
                                     only_pseudo=False):
    """Calculate magnetic moment matrix in LCAO basis.

    Parameters
    ----------
    kpt_u
        K-points
    bfs
        Basis functions object
    correction
        Correction object
    r_vG
        Grid point coordinates
    dM_vaii
        Atomic PAW corrections for magnetic moment
    only_pseudo
        If true, do not add PAW corrections

    Returns
    -------
    Magnetic moment matrix
    """
    Mstart = correction.Mstart
    Mstop = correction.Mstop
    mynao = Mstop - Mstart
    nao = bfs.Mmax

    assert bfs.Mstart == Mstart
    assert bfs.Mstop == Mstop

    M_vmM = np.zeros((3, mynao, nao), dtype=complex)
    rnabla_vmM = np.empty((3, mynao, nao), dtype=complex)

    for v in range(3):
        v1 = (v + 1) % 3
        v2 = (v + 2) % 3
        rnabla_vmM[:] = 0.0
        bfs.calculate_potential_matrix_derivative(r_vG[v], rnabla_vmM, 0)
        M_vmM[v1] += rnabla_vmM[v2]
        M_vmM[v2] -= rnabla_vmM[v1]

    if not only_pseudo:
        for kpt in kpt_u:
            assert kpt.k == 0
            for v in range(3):
                correction.calculate(kpt.q, dM_vaii[v], M_vmM[v],
                                     Mstart, Mstop)

    # The matrix should be real
    assert np.max(np.absolute(M_vmM.imag)) == 0.0
    M_vmM = M_vmM.real.copy()
    return -0.5 * M_vmM


def calculate_magnetic_moment_in_lcao(ksl, rho_mm, M_vmm):
    """Calculate magnetic moment in LCAO.

    Parameters
    ----------
    ksl
        Kohn-Sham Layouts object
    rho_mm
        Density matrix in LCAO basis
    M_vmm
        Magnetic moment matrix in LCAO basis

    Returns
    -------
    Magnetic moment vector
    """
    assert M_vmm.dtype == float
    mm_v = np.sum(rho_mm.imag * M_vmm, axis=(1, 2))
    if ksl.using_blacs:
        ksl.mmdescriptor.blacsgrid.comm.sum(mm_v)
    return mm_v


def get_origin_coordinates(atoms: Atoms,
                           origin: str,
                           origin_shift: Vector) -> np.ndarray:
    """Get origin coordinates.

    Parameters
    ----------
    atoms
        Atoms object
    origin
        See :class:`~gpaw.tddft.MagneticMomentWriter`
    origin_shift
        See :class:`~gpaw.tddft.MagneticMomentWriter`

    Returns
    -------
    Origin coordinates in atomic units
    """
    if origin == 'COM':
        origin_v = atoms.get_center_of_mass()
    elif origin == 'COC':
        origin_v = 0.5 * atoms.get_cell().sum(0)
    elif origin == 'zero':
        origin_v = np.zeros(3, dtype=float)
    else:
        raise ValueError('unknown origin')
    origin_v += np.asarray(origin_shift, dtype=float)
    return origin_v / Bohr


def parse_header(line: str) -> Tuple[str, int, dict]:
    """Parse header line.

    Example header line (keyword arguments as json):

        NameOfWriter[version=1](**{"arg1": "abc", ...})

    Parameters
    ----------
    line
        Header line

    Returns
    -------
    name
        Name
    version
        Version
    kwargs
        Keyword arguments

    Raises
    ------
    ValueError
        Line cannot be parsed
    """
    regexp = r"^(?P<name>\w+)\[version=(?P<ver>\d+)\]\(\*\*(?P<args>.*)\)$"
    m = re.match(regexp, line)
    if m is None:
        raise ValueError('unable parse header')
    name = m.group('name')
    version = int(m.group('ver'))
    try:
        kwargs = json.loads(m.group('args'))
    except json.decoder.JSONDecodeError:
        raise ValueError('unable parse keyword arguments')
    return name, version, kwargs


class MagneticMomentWriter(TDDFTObserver):
    """Observer for writing time-dependent magnetic moment data.

    The data is written in atomic units.

    The observer attaches to the TDDFT calculator during creation.

    Parameters
    ----------
    paw
        TDDFT calculator
    filename
        File for writing magnetic moment data
    origin
        Origin of the coordinate system used in calculation.
        Possible values:
        ``'COM'``: center of mass (default),
        ``'COC'``: center of cell,
        ``'zero'``: (0, 0, 0)
    origin_shift
        Vector in Å shifting the origin from the position defined
        by *origin*.
    dmat
        Density matrix object.
        Define this for LCAO calculations to avoid
        expensive recalculations of the density matrix.
    calculate_on_grid
        Parameter for testing.
        In LCAO mode, if true, calculation is performed on real-space grid.
        In fd mode, calculation is always performed on real-space grid
        and this parameter is neglected.
    only_pseudo
        Parameter for testing.
        If true, PAW corrections are neglected.
    interval
        Update interval. Value of 1 corresponds to evaluating and
        writing data after every propagation step.
    """
    version = 5

    def __init__(self, paw, filename: str, *,
                 origin: str = None,
                 origin_shift: Vector = None,
                 dmat: DensityMatrix = None,
                 calculate_on_grid: bool = None,
                 only_pseudo: bool = None,
                 interval: int = 1):
        TDDFTObserver.__init__(self, paw, interval)
        self.ioctx = IOContext()
        mode = paw.wfs.mode
        assert mode in ['fd', 'lcao'], 'unknown mode: {}'.format(mode)
        if paw.niter == 0:
            if origin is None:
                origin = 'COM'
            if origin_shift is None:
                origin_shift = [0., 0., 0.]
            if calculate_on_grid is None:
                calculate_on_grid = mode == 'fd'
            if only_pseudo is None:
                only_pseudo = False
            _kwargs = dict(origin=origin,
                           origin_shift=origin_shift,
                           calculate_on_grid=calculate_on_grid,
                           only_pseudo=only_pseudo)

            # Initialize
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='w')
            self._write_header(paw, _kwargs)
        else:
            if origin is not None:
                raise ValueError('Do not set origin in restart')
            if origin_shift is not None:
                raise ValueError('Do not set origin_shift in restart')
            if calculate_on_grid is not None:
                raise ValueError('Do not set calculate_on_grid in restart')
            if only_pseudo is not None:
                raise ValueError('Do not set only_pseudo in restart')

            # Read and continue
            _kwargs = self._read_header(filename)
            origin = _kwargs['origin']  # type: ignore
            origin_shift = _kwargs['origin_shift']  # type: ignore
            calculate_on_grid = _kwargs['calculate_on_grid']  # type: ignore
            only_pseudo = _kwargs['only_pseudo']  # type: ignore
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='a')

        atoms = paw.atoms
        gd = paw.wfs.gd
        self.timer = paw.timer

        assert isinstance(origin, str)
        assert isinstance(origin_shift, list)
        origin_v = get_origin_coordinates(atoms, origin, origin_shift)
        R_av = atoms.positions / Bohr - origin_v[np.newaxis, :]
        r_vG, _ = coordinates(gd, origin=origin_v)

        dM_vaii = calculate_magnetic_moment_atomic_corrections(
            R_av, paw.setups, paw.hamiltonian.dH_asp.partition)

        self.calculate_on_grid = calculate_on_grid
        if self.calculate_on_grid:
            self.only_pseudo = only_pseudo
            self.r_vG = r_vG
            self.dM_vaii = dM_vaii

            grad_v = []
            for v in range(3):
                grad_v.append(Gradient(gd, v, dtype=complex, n=2))
            self.grad_v = grad_v
        else:
            M_vmM = calculate_magnetic_moment_matrix(
                paw.wfs.kpt_u, paw.wfs.basis_functions,
                paw.wfs.atomic_correction, r_vG, dM_vaii,
                only_pseudo=only_pseudo)

            # TODO: All observers recalculate density matrix
            # unless dmat is given.
            # Calculator itself could have a density matrix object to avoid
            # this expensive recalculation in a clean way.
            if dmat is None:
                self.dmat = DensityMatrix(paw)
            else:
                self.dmat = dmat
            ksl = paw.wfs.ksl
            if ksl.using_blacs:
                self.M_vmm = ksl.distribute_overlap_matrix(M_vmM)
            else:
                gd.comm.sum(M_vmM)
                self.M_vmm = M_vmM

    def _write(self, line):
        self.fd.write(line)
        self.fd.flush()

    def _write_header(self, paw, kwargs):
        origin_v = get_origin_coordinates(
            paw.atoms, kwargs['origin'], kwargs['origin_shift'])
        lines = [f'{self.__class__.__name__}[version={self.version}]'
                 f'(**{json.dumps(kwargs)})',
                 'origin_v = [%.6f, %.6f, %.6f] Å' % tuple(origin_v * Bohr)]
        self._write('# ' + '\n# '.join(lines) + '\n')
        self._write(f'# {"time":>15} {"mmx":>17} {"mmy":>22} {"mmz":>22}\n')

    def _read_header(self, filename):
        with open(filename, 'r', encoding='utf-8') as fd:
            line = fd.readline()
        try:
            name, version, kwargs = parse_header(line[2:])
        except ValueError as e:
            raise ValueError(f'File {filename} cannot be parsed: {e}')
        if name != self.__class__.__name__:
            raise ValueError(f'File {filename} is not '
                             f'for {self.__class__.__name__}')
        if version != self.version:
            raise ValueError(f'File {filename} is not '
                             f'of version {self.version}')
        return kwargs

    def _write_init(self, paw):
        time = paw.time
        line = '# Start; Time = %.8lf\n' % time
        self._write(line)

    def _write_kick(self, paw):
        time = paw.time
        kick = paw.kick_strength
        line = '# Kick = [%22.12le, %22.12le, %22.12le]; ' % tuple(kick)
        line += 'Time = %.8lf\n' % time
        self._write(line)

    def _calculate_mm(self, paw):
        if self.calculate_on_grid:
            self.timer.start('Calculate magnetic moment on grid')
            mm_v = calculate_magnetic_moment_on_grid(
                paw.wfs, self.grad_v, self.r_vG, self.dM_vaii,
                only_pseudo=self.only_pseudo)
            self.timer.stop('Calculate magnetic moment on grid')
        else:
            self.timer.start('Calculate magnetic moment in LCAO')
            u = 0
            rho_mm = self.dmat.get_density_matrix((paw.niter, paw.action))[u]
            mm_v = calculate_magnetic_moment_in_lcao(
                paw.wfs.ksl, rho_mm, self.M_vmm)
            self.timer.stop('Calculate magnetic moment in LCAO')
        assert mm_v.shape == (3,)
        assert mm_v.dtype == float
        return mm_v

    def _write_mm(self, paw):
        time = paw.time
        mm_v = self._calculate_mm(paw)
        line = ('%20.8lf %22.12le %22.12le %22.12le\n'
                % (time, mm_v[0], mm_v[1], mm_v[2]))
        self._write(line)

    def _update(self, paw):
        if paw.action == 'init':
            self._write_init(paw)
        elif paw.action == 'kick':
            self._write_kick(paw)
        self._write_mm(paw)

    def __del__(self):
        self.ioctx.close()



from ase import Atoms
from ase.units import Bohr
from gpaw import GPAW
from gpaw.test import equal


def xc(name):
    return {'name': name, 'stencil': 1}


def test_xc_nonselfconsistent(in_tmp_dir):
    a = 7.5 * Bohr
    n = 16
    atoms = Atoms('He', [(0.0, 0.0, 0.0)], cell=(a, a, a), pbc=True)
    params = dict(mode='fd', gpts=(n, n, n), nbands=1)
    atoms.calc = GPAW(**params, xc=xc('PBE'))
    e1 = atoms.get_potential_energy()
    e1ref = atoms.calc.get_reference_energy()
    de12 = atoms.calc.get_xc_difference(xc('revPBE'))
    atoms.calc = GPAW(**params, xc=xc('revPBE'))
    e2 = atoms.get_potential_energy()
    e2ref = atoms.calc.get_reference_energy()
    de21 = atoms.calc.get_xc_difference(xc('PBE'))
    print(e1ref + e1 + de12 - (e2ref + e2))
    print(e1ref + e1 - (e2ref + e2 + de21))
    print(de12, de21)
    equal(e1ref + e1 + de12, e2ref + e2, 8e-4)
    equal(e1ref + e1, e2ref + e2 + de21, 3e-3)

    atoms.calc.write('revPBE.gpw')

    de21b = GPAW('revPBE.gpw').get_xc_difference(xc('PBE'))
    equal(de21, de21b, 9e-8)

    energy_tolerance = 0.0005
    equal(e1, -0.07904951, energy_tolerance)
    equal(e2, -0.08147563, energy_tolerance)



from ase import Atoms
from gpaw import GPAW
import pytest


@pytest.mark.mgga
def test_symm_mgga():
    a = 5.47
    b = a / 2
    si = Atoms('Si2',
               scaled_positions=[[0, 0, 0], [0.25, 0.25, 0.25]],
               cell=[[0, b, b], [b, 0, b], [b, b, 0]],
               pbc=True)
    k = 2
    xc = 'M06-L'
    energies = []
    for symmetry in [True, False]:
        si.calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                       kpts={'size': (k, k, k), 'gamma': True},
                       symmetry={} if symmetry else 'off',
                       xc=xc)
        e = si.get_potential_energy()
        energies.append(e)
    e1, e2 = energies
    assert e1 == pytest.approx(e2, abs=0.001)
